(*  
    Author:      René Thiemann 
                 Akihisa Yamada
    License:     BSD
*)
section \<open>Root Finder\<close>

text \<open>The theory combines the square free factorization algorithm with the algorithms
  which use the explicit closed forms for roots. There are instances for real and complex numbers.\<close>

theory Root_Finder
imports 
  Complex_Roots 
  Square_Free_Factorization
  Poly_Show
begin

definition complex_factorization :: 
  "complex list \<Rightarrow> complex poly list \<Rightarrow> complex poly \<Rightarrow> complex \<times> (complex poly \<times> nat) list" where
  "complex_factorization = root_finder_samples_factorization croots_no_sqrt"

definition real_factorization :: 
  "real list \<Rightarrow> real poly list \<Rightarrow> real poly \<Rightarrow> real \<times> (real poly \<times> nat) list" where
  "real_factorization = root_finder_samples_factorization rroots_no_sqrt"

lemma complex_factorization: "square_free_factorization p (complex_factorization samples psamples p)"
  unfolding complex_factorization_def by (rule root_finder_samples_factorization)

lemma real_factorization: "square_free_factorization p (real_factorization samples psamples p)"
  unfolding real_factorization_def by (rule root_finder_samples_factorization)

definition monic_linear_factors :: "'a :: field poly \<Rightarrow> nat \<Rightarrow> ('a \<times> 'a list)" where
  "monic_linear_factors p n \<equiv> 
    let 
      rt = roots1 p;
      c  = coeff (p div [: -rt, 1 :]) 0
    in (c ^ n, replicate n rt)"

lemma monic_linear_factors: assumes deg: "degree p = 1"
  and mon: "monic_linear_factors p n = (c, rt)"
  shows "p^n = smult c (\<Prod> a \<leftarrow> rt. [:- a, 1:])"
proof -
  let ?rt = "roots1 p"
  let ?q = "[: - ?rt, 1 :]"
  from roots1[OF deg] have "poly p ?rt = 0" by auto
  from this[unfolded poly_eq_0_iff_dvd] have "?q dvd p" .
  then obtain r where p: "p = ?q * r" unfolding dvd_def by auto
  with deg have qr: "?q \<noteq> 0" "r \<noteq> 0" by auto
  from p qr have div: "p div ?q = r" by (metis div_mult_self1_is_id)
  note mon = mon[unfolded monic_linear_factors_def Let_def div]
  from degree_mult_eq[OF qr, folded p deg[symmetric]] have "degree r = 0" by auto
  from degree0_coeffs[OF this] obtain d where "r = [: d :]" by auto
  hence r: "r = [: coeff r 0 :]" by auto
  have "p ^ n = (r * ?q)^n" using p by (auto simp: ac_simps)
  also have "r * ?q = (smult (coeff r 0) ?q)" using r
    by (metis add.right_neutral mult_pCons_left mult_zero_left pCons_0_0)
  also have "\<dots> ^ n = smult ((coeff r 0) ^ n) (?q ^ n)"
    by (rule smult_power)
  also have "(coeff r 0) ^ n = c" using mon by simp
  also have "?q ^ n = (\<Prod> a \<leftarrow> replicate n ?rt. [:- a, 1:])" 
    by (induct n, auto)
  also have "replicate n ?rt = rt" using mon by simp
  finally show ?thesis .
qed

fun linear_root_extractor_main :: "'a :: field \<Rightarrow> ('a poly \<times> nat) list \<Rightarrow> ('a \<times> ('a list))option" where
  "linear_root_extractor_main c [] = Some (c,[])"
| "linear_root_extractor_main c ((p,i) # bs) = (if degree p = 1 then
     (case monic_linear_factors p (Suc i) of (d, xs) \<Rightarrow>
     (case linear_root_extractor_main (c * d) bs of
        Some (e,ys) \<Rightarrow> Some (e, xs @ ys)
      | None \<Rightarrow> None)) else None)"
      
definition linear_root_extractor :: "'a :: field \<times> ('a poly \<times> nat) list \<Rightarrow> ('a \<times> ('a list))option" where
  "linear_root_extractor c_bs \<equiv> case c_bs of (c,bs) \<Rightarrow>
    linear_root_extractor_main c bs"

lemma linear_root_extractor: assumes sff: "square_free_factorization p c_bs"
  and lre: "linear_root_extractor c_bs = Some (c,xs)"
  shows "p = smult c (\<Prod> a \<leftarrow> xs. [:- a, 1:])"
proof -
  obtain d bs where c_bs: "c_bs = (d,bs)" by force
  from lre[unfolded linear_root_extractor_def c_bs split]
  have lin: "linear_root_extractor_main d bs = Some (c, xs)" .
  from square_free_factorizationD[OF sff[unfolded c_bs]]
  have p: "p = smult d (\<Prod>(a, i)\<in>set bs. a ^ Suc i)" and dist: "distinct bs" by blast+
  from lin dist show ?thesis unfolding p
  proof (induct bs arbitrary: d c xs)
    case (Cons pi bs d c xs)
    note IH = Cons(1)
    obtain p i where pi: "pi = (p,i)" by force
    obtain e ys where mon: "monic_linear_factors p (Suc i) = (e,ys)" by force
    let ?rec = "linear_root_extractor_main (d * e) bs"
    note Some = Cons(2)[unfolded pi linear_root_extractor_main.simps mon split]
    obtain rec where rec: "?rec = Some rec" using Some by (auto split: option.splits)
    from Some have p: "degree p = 1" by (auto split: if_splits)
    from Some[unfolded rec] p obtain zs where rec2: "rec = (c,zs)" and xs: "xs = ys @ zs"
      by (cases rec, auto)
    from IH[OF rec[unfolded rec2]] Cons(3)
    have IH: "smult (d * e) (\<Prod>(a, i)\<in>set bs. a ^ Suc i) = smult c (\<Prod>a\<leftarrow>zs. [:- a, 1:])" by auto
    from monic_linear_factors[OF p mon] have p: "p ^ Suc i = smult e (\<Prod>a\<leftarrow>ys. [:- a, 1:])" by auto    
    have "smult d (\<Prod>(a, i)\<in>set (pi # bs). a ^ Suc i) = 
      smult d (p ^ Suc i * (\<Prod>(a, i)\<in>set bs. a ^ Suc i))"
      by (subst setprod.remove[of _ pi], insert Cons(3), auto simp: pi)
    also have "\<dots> = (smult (d * e) (\<Prod>(a, i)\<in>set bs. a ^ Suc i)) * (\<Prod>a\<leftarrow>ys. [:- a, 1:])"
      unfolding p by (simp add: ac_simps)
    also have "\<dots> = smult c ((\<Prod>a\<leftarrow>ys. [:- a, 1:]) * (\<Prod>a\<leftarrow>zs. [:- a, 1:]))"
      unfolding IH by (simp add: ac_simps)
    also have "((\<Prod>a\<leftarrow>ys. [:- a, 1:]) * (\<Prod>a\<leftarrow>zs. [:- a, 1:])) = ((\<Prod>a\<leftarrow>xs. [:- a, 1:]))"
      unfolding xs by simp
    finally show ?case .
  qed simp
qed

definition full_monic_poly_root_extractor 
  :: "('a :: field_char_0 poly \<Rightarrow> 'a list) \<Rightarrow> 'a list \<Rightarrow> 'a poly list \<Rightarrow> 'a poly \<Rightarrow> 'a list option" where
  "full_monic_poly_root_extractor root_finder samples psamples p \<equiv> 
     case linear_root_extractor (root_finder_samples_factorization root_finder samples psamples p)
     of Some (i,xs) \<Rightarrow> Some xs
       | None \<Rightarrow> None"

lemma full_monic_poly_root_extractor: 
  assumes full: "full_monic_poly_root_extractor root_finder samples psamples p = Some xs"
  and mon: "monic p"
  shows "p = (\<Prod> a \<leftarrow> xs. [:- a, 1:])"
proof -
  let ?rt = "root_finder_samples_factorization root_finder samples psamples p"  
  let ?lin = "linear_root_extractor ?rt"
  note full = full[unfolded full_monic_poly_root_extractor_def]
  obtain res where lin: "?lin = Some res" using full by (auto split: option.splits)
  from full[unfolded lin] obtain i where res: "res = (i,xs)" by (cases res, auto)
  from linear_root_extractor[OF root_finder_samples_factorization lin[unfolded res]]
  have p: "p = smult i (\<Prod>a\<leftarrow>xs. [:- a, 1:])" by auto
  have "monic (\<Prod>a\<leftarrow>xs. [:- a, 1:])"
    by (rule monic_listprod, auto)
  with mon[unfolded p] have "i = 1"
    by (metis (no_types, lifting) degree_smult_eq mult.comm_neutral smult.rep_eq smult_eq_0_iff)
  with p show ?thesis by simp
qed 

definition complex_root_extractor :: 
  "complex list \<Rightarrow> complex poly list \<Rightarrow> complex poly \<Rightarrow> complex list option" where
  "complex_root_extractor = full_monic_poly_root_extractor croots_no_sqrt"

definition real_root_extractor :: 
  "real list \<Rightarrow> real poly list \<Rightarrow> real poly \<Rightarrow> real list option" where
  "real_root_extractor = full_monic_poly_root_extractor rroots_no_sqrt"

lemma complex_root_extractor: assumes ext: "complex_root_extractor samples psamples p = Some xs"
  and mon: "monic p"
  shows "p = (\<Prod> a \<leftarrow> xs. [:- a, 1:])"
  by (rule full_monic_poly_root_extractor[OF ext[unfolded complex_root_extractor_def] mon])

lemma real_root_extractor: assumes ext: "real_root_extractor samples psamples p = Some xs"
  and mon: "monic p"
  shows "p = (\<Prod> a \<leftarrow> xs. [:- a, 1:])"
  by (rule full_monic_poly_root_extractor[OF ext[unfolded real_root_extractor_def] mon])


end