(*  
    Author:      René Thiemann 
                 Akihisa Yamada
    License:     BSD
*)
theory Schur_Decomposition_Impl
imports 
  Root_Finder
  Schur_Decomposition
  "$AFP/Real_Impl/Real_Unique_Impl"
begin

datatype sqrt_analysis = No_Sqrt | Single_Sqrt nat | Multiple_Sqrt

fun sqrt_combine :: "sqrt_analysis \<Rightarrow> sqrt_analysis \<Rightarrow> sqrt_analysis" where
  "sqrt_combine (Single_Sqrt a) (Single_Sqrt b) = (if a = b then Single_Sqrt a else Multiple_Sqrt)"
| "sqrt_combine No_Sqrt y = y"
| "sqrt_combine x No_Sqrt = x"
| "sqrt_combine Multiple_Sqrt y = Multiple_Sqrt"
| "sqrt_combine x Multiple_Sqrt = Multiple_Sqrt"

fun sqrt_list_combine :: "sqrt_analysis list \<Rightarrow> sqrt_analysis" where
  "sqrt_list_combine [] = No_Sqrt"
| "sqrt_list_combine (a # as) = sqrt_combine a (sqrt_list_combine as)"

definition real_sqrts :: "real \<Rightarrow> sqrt_analysis" where
  "real_sqrts x = (if (\<exists> u. x = real_of_u u) then
    (let u = (SOME u. (x = real_of_u u));
        b = mau_base u
     in if b = 0 then No_Sqrt else Single_Sqrt b) else Multiple_Sqrt)"

definition complex_sqrts :: "complex \<Rightarrow> sqrt_analysis" where
  "complex_sqrts x = (sqrt_combine (real_sqrts (Re x)) (real_sqrts (Im x)))"

definition reals_compatible_analysis :: "real list \<Rightarrow> bool" where
  "reals_compatible_analysis xs = (sqrt_list_combine (map real_sqrts xs) \<noteq> Multiple_Sqrt)"

definition complex_compatible_analysis :: "complex list \<Rightarrow> bool" where
  "complex_compatible_analysis xs = (sqrt_list_combine (map complex_sqrts xs) \<noteq> Multiple_Sqrt)"

lemma real_sqrt_impl[code]: "real_sqrts (real_of_u r) = (let b = mau_base r in
    if b = 0 then No_Sqrt else Single_Sqrt b)" (is "_ = ?impl")
proof -
  let ?x = "real_of_u r"
  def u \<equiv> "SOME u. ?x = real_of_u u"
  have "real_of_u u = real_of_u r"
    unfolding u_def by auto
  from this[unfolded real_of_u_inj]
  have ur: "u = r" by auto
  have "real_sqrts ?x = (if mau_base u = 0 then No_Sqrt else Single_Sqrt (mau_base u))"
    unfolding real_sqrts_def Let_def u_def[symmetric] by simp
  also have "\<dots> = ?impl" unfolding ur Let_def ..
  finally show ?thesis .
qed

definition schur_approx_complex :: "complex list \<Rightarrow> complex poly list \<Rightarrow> complex mat \<Rightarrow> complex mat option" where
  "schur_approx_complex samples psamples A \<equiv> 
     case complex_root_extractor samples psamples (char_poly A) of
       Some xs \<Rightarrow> if complex_compatible_analysis xs then Some (schur_upper_triangular A xs) else None
     | None \<Rightarrow> None"

lemma schur_approx_complex: assumes schur: "schur_approx_complex samples psamples A = Some B"
  and A: "A \<in> carrier\<^sub>m n n"
  shows "B \<in> carrier\<^sub>m n n"
    "upper_triangular B" "mat_similar A B" 
proof -
  note schur = schur[unfolded schur_approx_complex_def]
  let ?cre = "complex_root_extractor samples psamples (char_poly A)"
  from schur obtain xs where cre: "?cre = Some xs" and B: "B = schur_upper_triangular A xs"
    by (auto split: option.splits if_splits)
  from complex_root_extractor[OF cre] degree_monic_char_poly[OF A]
  have "char_poly A = (\<Prod>a\<leftarrow>xs. [:- a, 1:])" by auto
  from schur_upper_triangular[OF A this, folded B]
  show "B \<in> carrier\<^sub>m n n" "upper_triangular B" "mat_similar A B" .
qed
  


end